"""Constants used for property tests."""
# Minimum absolute value of floats.
FLOAT_MIN = 1e-3

# Absolute tolerance for math.isclose.
ATOL = 1e-4

# Min and max dimensions of arrays.
DIM_MIN, DIM_MAX = 2, 10
