"""Package containing spatial objects."""
from skspatial.objects.circle import Circle
from skspatial.objects.cylinder import Cylinder
from skspatial.objects.line import Line
from skspatial.objects.plane import Plane
from skspatial.objects.point import Point
from skspatial.objects.points import Points
from skspatial.objects.sphere import Sphere
from skspatial.objects.triangle import Triangle
from skspatial.objects.vector import Vector

__all__ = ['Circle', 'Cylinder', 'Line', 'Plane', 'Point', 'Points', 'Sphere', 'Triangle', 'Vector']
